/*  PRESTO: SLOCK.c
 *
 *  The purpose of this module is to provide out-of-line procedure
 *  definitions for S_LOCK, etc., since C++ doesn't understand
 *  asm-functions.  The preferred solution to this problem, however,
 *  is to use "CC +h/usr/include/Presto/asmdefs.h file.c";
 *  this causes the asm-function definitions to be slipped into
 *  file..c before cc is called.
 *
 *  (Except that this relies on a non-standard C++ compiler option)
 */

#ifdef i386

/* $Copyright:	$
 * Copyright (c) 1984, 1985, 1986, 1987 Sequent Computer Systems, Inc.
 * All rights reserved
 *  
 * This software is furnished under a license and may be used
 * only in accordance with the terms of that license and with the
 * inclusion of the above copyright notice.   This software may not
 * be provided or otherwise made available to, or used by, any
 * other person.  No title to or ownership of the software is
 * hereby transferred.
 */

/* $Header: parallel.h 1.15 87/11/03 $
 *
 * parallel.h
 *	Definitions for use in parallel, shared-memory programs.
 *
 */

/*
 * A "lock" is a byte of memory, initialized to zero (unlocked).
 */
typedef unsigned char   slock_t;

#define	L_UNLOCKED	0
#define	L_LOCKED	1

/*
 * Was a conditional lock request granted (L_SUCCESS) or denied (L_FAILED)
 */

#define L_FAILED	0
#define L_SUCCESS	1

/*
 * LOCK() UNLOCK(), CLOCK(), INIT_LOCK, and WAIT_BARRIER functions.
 * These are implemented as "asm functions", and will produce very good code.
 *
 * Need to use the -i flag with the optimizer when these inline functions are
 * used. (cc -O -i).   This is to prevent the optimizer from removing
 * "redundant" moves (which aren't really redundant in this case)
 * in the spin loops.
 */

asm void LOCK(laddr)
{
%reg laddr; lab loop, spin, done;
/PEEPOFF
loop:	movb	$L_LOCKED, %al
	xchgb	%al, (laddr)		/* an atomic "test and set" */
	cmpb	$L_UNLOCKED, %al
	je	done			/* if equal, we got the lock */
spin:	cmpb	$L_UNLOCKED, (laddr)	/* spin in cache until unlocked */
	je	loop			/* then, try get lock again */
	jmp	spistdone:
/PEEPON
%mem laddr; lab loop, spin, done;
/PEEPOFF
loop:	movb	$L_LOCKED, %al
	movl	laddr, %ecx
	xchgb	%al, (%ecx)		/* an atomic "test and set" */
	cmpb	$L_UNLOCKED, %al
	je	done			/* if equal, we got the lock */
spin:	cmpb	$L_UNLOCKED, (%ecx)	/* spin in cache until unlocked */
	je	loop			/* then, try get lock again */
	jmp	spin
done:
/PEEPON
}

asm void UNLOCK(laddr)
{
%reg laddr;
	movb	$L_UNLOCKED, %al
	xchgb	%al, (laddr)		/* clear lock, "atomically" */
%mem laddr;
	movb	$L_UNLOCKED, %al
	movl	laddr, %ecx
	xchgb	%al, (%ecx)		/* clear lock, "atomically" */
}

asm CLOCK(laddr)
{
%reg laddr; lab done;
	movb	$L_LOCKED, %dl
	movl	$L_SUCCESS, %eax
	xchgb	%dl, (laddr)		/* test and set lock */
	cmpb	$L_UNLOCKED, %dl	/* if we got it, return success */
	je	done
	movl	$L_FAILED, %eax		/* else return failure */
done:
%mem laddr; lab done;
	movb	$L_LOCKED, %dl
	movl	laddr, %ecx
	movl	$L_SUCCESS, %eax
	xchgb	%dl, (%ecx)		/* test and set lock */
	cmpb	$L_UNLOCKED, %dl	/* if we got it, return success */
	je	done
	movl	$L_FAILED, %eax		/* else return failure */
done:
}

/*
 * In line barrier initialization macro
 *
 * Note, there's no need to initialize either b_useno (due to its manner
 * of useage, or b_mutex (due to its non-usage).
 */

#define INIT_BARRIER(b, l)	((b)->b_limit = (b)->b_count = (l))

/*
 * In line barrier wait procedure.
 *
 * Define offsets of fields in barrier, for use in assembly language.
 */

#define MUTEX	0
#define COUNT	1
#define LIMIT	2
#define USENO	3

asm void WAIT_BARRIER(baddr)
{
%reg baddr; lab spin;
/PEEPOFF
	movb	USENO(baddr), %al	/* save old barrier useno value */
	lock				/* assert lock on next instruction */
	decb	COUNT(baddr)		/* atomic decrement of barrier count */
	jne	spin			/* if not all checked in, go spin */
	movb	LIMIT(baddr), %dl
	movb	%dl, COUNT(baddr) 	/* all checked in! re-init barrier */
	incb	USENO(baddr)		/* and free all the spinners */
spin:	cmpb	USENO(baddr), %al	/* spin here until all are checked in */
	je	spin
/PEEPON
%mem baddr; lab spin;
/PEEPOFF
	movl	baddr, %ecx		/* load barrier pointer into register */
	movb	USENO(%ecx), %al	/* save old barrier useno value */
	lock				/* assert lock on next instruction */
	decb	COUNT(%ecx)		/* atomic decrement of barrier count */
	jne 	spin			/* if not all checked in, go spin */
	movb	LIMIT(%ecx), %dl
	movb	%dl, COUNT(%ecx)	/* all checked in! re-init barrier */
	incb	USENO(%ecx)		/* and free all the spinners */
spin:	cmpb	USENO(%ecx), %al	/* spin here until all are checked in */
	je	spin
/PEEPON
}


void
#ifdef __STDC__
S_INIT_LOCK(slock_t *lock)
#else
S_INIT_LOCK(lock)
	slock_t	*lock;
#endif
{
	UNLOCK(lock);
}

void
#ifdef __STDC__
S_LOCK(register slock_t *lock)
#else
S_LOCK(lock)
	register slock_t	*lock;
#endif
{
	LOCK(lock);
}

void
#ifdef __STDC__
S_UNLOCK(slock_t *lock)
#else
S_UNLOCK(lock)
	slock_t	*lock;
#endif
{
	UNLOCK(lock);
}

void
#ifdef __STDC__
S_CLOCK(slock_t *lock)
#else
S_CLOCK(lock)
	slock_t	*lock;
#endif
{
	CLOCK(lock);
}

void
#ifdef __STDC__
S_WAIT_BARRIER(slock_t *lock)
#else
S_WAIT_BARRIER(lock)
	slock_t	*lock;
#endif
{
    WAIT_BARRIER(lock);
}

#endif /* i386 */
